class DeSearchSpace:
    def __init__():
        pass

class DeMainNet:
    def __init__():
        pass

    @classmethod
    def create_netblock_list_from_str(cls, struct_str, no_create=False, **kwargs):
        pass